/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.7 $
 *********************************************************************/

#ifndef _COORDINATES_H
#define _COORDINATES_H

namespace prapi
{
	/**
	 * A class for storing 2D coordinates. This is mainly a convenience
	 * class for bundling two integer values into an easily-accessible
	 * package.
	 **/
	template <class T> class Point
	{
	public:
		/**
		 * Initialize a point with the given coordinates.
		 **/
		Point(T xOrigin=0,T yOrigin=0) { x=xOrigin; y=yOrigin; }
		/**
		 * Copy a point.
		 **/
		Point(const Point& other) { copy(other); }

		/**
		 * Make a copy of a point.
		 **/
		Point& operator= (const Point& other) { copy(other); return *this; }

		/**
		 * The x coordinate.
		 **/
		T x;
		/**
		 * The y coordinate.
		 **/
		T y;

		bool operator== (const Point& other) { return (x == other.x)&&(y == other.y);}
		Point& operator*= (const Point& other) { x *= other.x; y*=other.y; return *this; }
		Point& operator/= (const Point& other) { x /= other.x; y/=other.y; return *this; }
		Point& operator+= (const Point& other) { x += other.x; y+=other.y; return *this; }
		Point& operator-= (const Point& other) { x -= other.x; y-=other.y; return *this; }

		Point& operator*= (T value) { x *= value; y*=value; return *this; }
		Point& operator/= (T value) { x /= value; y/=value; return *this; }
		Point& operator+= (T value) { x += value; y+=value; return *this; }
		Point& operator-= (T value) { x -= value; y-=value; return *this; }
		
	private:
		void copy(const Point& other) { x=other.x; y=other.y; }
	};

	typedef Point<int> IntegerPoint;
	typedef Point<float> FloatPoint;
	typedef Point<double> DoublePoint;
	
	template <class T> bool operator== (const Point<T>& p1, const Point<T>& p2)
	{
		return ((p1.x==p2.x)&&(p1.y==p2.y));
	}
	
	template <class T> Point<T> operator* (const Point<T>& p1, const Point<T>& p2)
	{
		Point<T> result(p1);
		result.x *= p2.x;
		result.y *= p2.y;
		return result;
	}
	template <class T> Point<T> operator/ (const Point<T>& p1, const Point<T>& p2)
	{
		Point<T> result(p1);
		result.x /= p2.x;
		result.y /= p2.y;
		return result;
	}
	template <class T> Point<T> operator+ (const Point<T>& p1, const Point<T>& p2)
	{
		Point<T> result(p1);
		result.x += p2.x;
		result.y += p2.y;
		return result;
	}
	template <class T> Point<T> operator- (const Point<T>& p1, const Point<T>& p2)
	{
		Point<T> result(p1);
		result.x -= p2.x;
		result.y -= p2.y;
		return result;
	}
	template <class T> Point<T> operator* (const Point<T>& p1, T value)
	{
		Point<T> result(p1);
		result.x *= value;
		result.y *= value;
		return result;
	}
	template <class T> Point<T> operator/ (const Point<T>& p1, T value)
	{
		Point<T> result(p1);
		result.x /= value;
		result.y /= value;
		return result;
	}
	template <class T> Point<T> operator+ (const Point<T>& p1, T value)
	{
		Point<T> result(p1);
		result.x += value;
		result.y += value;
		return result;
	}
	template <class T> Point<T> operator- (const Point<T>& p1, T value)
	{
		Point<T> result(p1);
		result.x -= value;
		result.y -= value;
		return result;
	}
}

#endif
