/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.10 $
 *********************************************************************/

#ifndef _FEATURESELECTOR_H
#define _FEATURESELECTOR_H

#include <Event.h>
#include "ga/GeneticEngine.h"
#include <Heap.h>
#include "ConfusionMatrix.h"
#include "Classifier.h"
#include "ProximityMatrix.h"

namespace prapi
{
	typedef prapi::ga::EvolutionEvent<char> EvolutionEvent;

	/**
	 * An optimization event is fired for each completed level
	 * (generation) in the optimization cycle.
	 **/
	class OptimizationEvent
	{
	public:
		/**
		 * Create a new optimization event.
		 *
		 * @param i the index of the current generation
		 * @param r the best result obtained so far
		 * @param the features that got the best result
		 **/
		OptimizationEvent(int i, double r, const util::List<int>& bestFeatures) :
			generationIndex(i), result(r), features(bestFeatures) {}

		/**
		 * The index of the current generation (optimization level).
		 **/
		int generationIndex;
		/**
		 * The fitness of the best individual, i.e. the best
		 * classification accuracy (minus length penalties) on this level.
		 **/
		double result;
		/**
		 * The features that got the best result.
		 **/
		util::List<int> features;
	};
	
	/**
	 * FeatureSelector is a class for finding an optimal subset of
	 * features. Typically, optimization samples with a great amount of
	 * features are created. In a subsequent step, 'useless' features
	 * are removed. The remaining features can then be used in
	 * classification. FeatureSelector uses a genetic algorithm in
	 * finding an optimal subset of patterns. It can also perform
	 * special operations needed for statistical features.
	 * <p>
	 *
	 * It is also possible to use the non-randomized version of a
	 * genetic algorithm for the optimization process. In the
	 * non-randomized case, beam search is utilized.
	 * <p>
	 *
	 * One can observe the optimization process by adding an event
	 * listener to a feature selector. After each search level
	 * (generation), all registered listeners are informed with an
	 * OptimizationEvent. In this point, one has the possibility of
	 * stopping the optimization with the stop() method. It is thus
	 * possible to impose optimization constraints other than iteration
	 * count and result threshold.
	 *
	 * @see OptimizationEvent
	 **/
	template <class T=double, class I=std::string, class C=int> class FeatureSelector :
		public util::EventSource<OptimizationEvent>,
		public util::EventListener<EvolutionEvent>,
		public prapi::ga::FitnessCalculator<char>,
		public prapi::ga::Mutator<char>
	{
	public:
		/**
		 * Create a new FeatureSelector that optimizes a feature set using
		 * the given samples. In non-randomized search (beam search),
		 * generation count is read 'search depth' and population size is
		 * 'beam width'.
		 *
		 * @param samples a list of optimization samples
		 * @param classCount the number of different classes in the sample set
		 * @param resultThreshold the value over which classification result
		 *        must rise before the evolution is stopped
		 * @param generationCount maximum number of generations to produce
		 *        (-1 disables)
		 * @param populationSize the number of individuals in each generation
		 * @param accumulate if statistical features are used, it is sometimes
		 *        desirable to be able to collect the values of disabled
		 *        features to an additional distribution bin. If this flag
		 *        is true, all disabled features are summed up and used as
		 *        an additional feature.
		 **/
		FeatureSelector(util::List<Sample<T,I,C> >& samples,
										int classCount,
										double resultThreshold,
										int generationCount = -1,
										int populationSize = 10,
										bool accumulate = false) :
			prapi::ga::Mutator<char>(0.1),
			_lstSamples(samples),
			_iClassCount(classCount),
			_dResultThreshold(resultThreshold),
			_iGenerationCount(generationCount),
			_iPopulationSize(populationSize),
			_bAccumulate(accumulate),
			_bStopped(false),
			_dLengthPenalty(0) {}

		virtual ~FeatureSelector() {}
		
		/**
		 * This method is called by a genetic algorithm each time a new
		 * generation is created. FeatureSelector keeps track of the
		 * current state of the optimization process by checking the
		 * results after each generation.
		 **/
		void eventOccured(util::EventSource<EvolutionEvent>* source, const EvolutionEvent& event);

		/**
		 * Calculate the fitness of a population. This method goes through
		 * all individuals in a population, generates a sample set for
		 * each individual (an individual represents the enabled
		 * features), and classifies the formed sample set. The fitness of
		 * an individual is the rate of correct classifications.
		 **/
		virtual void calculateFitness(util::List<prapi::ga::Individual<char> >& population)
			throw (ClassificationException&,util::InvalidArgumentException&);

		/**
		 * Mutate a single gene in an individual.
		 **/
		char mutate(char value) { return char(drand48()+0.5); }

		/**
		 * Get the classification result for a set of samples. The
		 * returned value is used as a fitness for an individual in a
		 * population. The default implementation uses a kNN classifier
		 * with k=3 and an Euclidean proximity measure. The classification
		 * result is obtained using a leave-one-out test. One may want to
		 * override this method to perform a different type of
		 * classification.
		 *
		 * @param samples the samples to be classified
		 * @return the fraction of correct classifications, [0,1]
		 **/
		virtual double getClassificationResult(util::List<Sample<T,I,C> >& samples)
			throw (ClassificationException&,InvalidArgumentException&);

		/**
		 * Optimize a feature set. A genetic algorithm is run either a
		 * predefined number of iterations or until a threshold for
		 * classification result is exceeded.
		 *
		 * @return the remaining feature sets in a list. The size of the
		 * list is equal to the size of the initial population. Each
		 * element in this list stores a variable number of indices to
		 * enabled features. Thus, result[0][5] references the sixth
		 * enabled feature in the first feature set. The returned list is
		 * sorted in descending order according to classification
		 * accuracy.
		 * @exception ClassificationException& if something goes kablooie.
		 **/
		util::List<util::List<int> > optimize() throw (ClassificationException&,util::InvalidArgumentException&)
		{ return optimize(NULL); }
		/**
		 * Optimize a feature set. If you want to continue an interrupted
		 * optimization, an initial population may be given. In the
		 * population, each gene in an individual is either 0 (off) or -1
		 * (on). The number of genes equals the number of possible
		 * features. For example, an individual having the genome (-1 0 0
		 * -1) represents a set of two enabled features from a maximum
		 * number of four features. Optimize() changes the contents of
		 * population during the optimization process.
		 **/
		util::List<util::List<int> > optimize(util::List<prapi::ga::Individual<char> >& population)
			throw (ClassificationException&,util::InvalidArgumentException&)
		{ return optimize(&population); }

		/**
		 * Optimize a feature set with beam search.
		 *
		 * @see #optimize()
		 **/
		util::List<util::List<int> > beamOptimize() throw (ClassificationException&,util::InvalidArgumentException&)
		{ return beamOptimize(NULL); }
		/**
		 * Optimize a feature set with beam search.
		 *
		 * @see #optimize()
		 **/
		util::List<util::List<int> > beamOptimize(util::List<prapi::ga::Individual<char> >& population)
			throw (ClassificationException&,util::InvalidArgumentException&)
		{ return beamOptimize(&population); }

		/**
		 * Create a new sample set that contains the samples in
		 * <i>samples</i> with only the indicated features collected into
		 * the feature vectors. If the accumulation flag is set, then an
		 * additional feature containing the sum of all disabled features
		 * is added to the feature vectors.
		 *
		 * @param samples the samples
		 * @param selectedFeatures the indices of enabled features
		 *
		 * @return a list of samples with the indicated features enabled
		 **/
		util::List<Sample<T,I,C> > createSampleSet(util::List<Sample<T,I,C> >& samples,
																							 util::List<int>& selectedFeatures);

		/**
		 * Set the result threshold.
		 **/
		void setResultThreshold(double threshold) { _dResultThreshold = threshold; }
		/**
		 * Get the result threshold.
		 **/
		double getResultThreshold() { return _dResultThreshold; }

		/**
		 * Set the maximum number of generations to produce. For beam
		 * search, this value determines the number of selected features
		 * in the result set. Set to -1 to disable the generation (search
		 * level) limit.
		 **/
		void setGenerationCount(int cnt) { _iGenerationCount = cnt; }
		/**
		 * Get the generation count.
		 **/
		int getGenerationCount() { return _iGenerationCount; }

		/**
		 * Set the number of classes in the data.
		 **/
		void setClassCount(int cnt) { _iClassCount = cnt; }
		/**
		 * Get the number of classes in the data.
		 **/
		int getClassCount() { return _iClassCount; }

		/**
		 * Set the number of individuals in each generation. For beam
		 * search, this is the width of the beam.
		 **/
		void setPopulationSize(int size) { _iPopulationSize = size; }
		/**
		 * Get the size of the population, i.e. the beam width.
		 **/
		int getPopulationSize() { return _iPopulationSize; }		

		/**
		 * Set the accumulation flag.
		 **/
		void setAccumulate(bool accumulate) { _bAccumulate = accumulate; }
		/**
		 * Get the accumulation flag.
		 **/
		bool getAccumulate() { return _bAccumulate; }

		/**
		 * Stop the optimization. The optimization will be finished as
		 * soon as possible, in most cases after the next generation
		 * (search) level has been generated.
		 **/
		void stop() { _bStopped = true; }

		/**
		 * Set the length penalty. If the length penalty is set to a value
		 * other than zero, each feature in a feature vector is 'punished'
		 * with the given value.
		 **/
		void setLengthPenalty(double value) { _dLengthPenalty = value; }
		/**
		 * Get the length penalty.
		 **/
		double getLengthPenalty() { return _dLengthPenalty; }
		/**
		 * Set the length penalty. If the length penalty is set to a value
		 * other than zero, each feature in a feature vector is 'punished'
		 * with the given value.
		 **/
		double& lengthPenalty() { return _dLengthPenalty; }

		/**
		 * Check if the given sequence of number is 'saturated'. A
		 * sequence of number is saturated if it does not grow
		 * sufficiently. The purpose of this method is to provide an easy
		 * way to check when to stop an optimization process. If the
		 * obtained classification results do not get better fast enough,
		 * it may be wise to stop the process.
		 *
		 * @param values a sequence of obtained classification results.
		 * The values should be ordered so that values[0] referes to the
		 * oldest entry.
		 *
		 * @param threshold a threshold for deciding what is sufficient
		 * increase in classification accuracy. If the accuracy does not,
		 * on average, increase more than the indicated threshold on each
		 * round, the process is saturated.
		 **/
		static bool isSaturated(const List<double>& values, double threshold);
		
	private:
		util::List<Sample<T,I,C> >& _lstSamples;
		int _iClassCount;
		double _dResultThreshold;
		int _iGenerationCount, _iPopulationSize;
		bool _bAccumulate;
		bool _bStopped;
		double _dLengthPenalty;

		util::List<util::List<int> > beamOptimize(util::List<prapi::ga::Individual<char> >* population)
			throw (ClassificationException&,util::InvalidArgumentException&);

		util::List<util::List<int> > optimize(util::List<prapi::ga::Individual<char> >* population)
			throw (ClassificationException&,util::InvalidArgumentException&);

		void createNewLevel(util::List<prapi::ga::Individual<char> >& population)
			throw (ClassificationException&,util::InvalidArgumentException&);

		util::List<util::List<int> > createSets(const util::List<prapi::ga::Individual<char> >& population);

		void calculateFitness(prapi::ga::Individual<char>& individual)
			throw (ClassificationException&);
	};

	template <class T, class I, class C> double FeatureSelector<T,I,C>::getClassificationResult(util::List<Sample<T,I,C> >& samples)
		throw (ClassificationException&,InvalidArgumentException&)
	{
		kNNClassifier<T,I,C> knn(samples,
												 new EuclideanDistance<T>,
												 _iClassCount,
												 3);
		knn.leaveOneOut();
		ConfusionMatrix mat(samples, _iClassCount);
		return 1-mat.getError();
	}

	template <class T, class I, class C>
	util::List<Sample<T,I,C> > FeatureSelector<T,I,C>::createSampleSet(util::List<Sample<T,I,C> >& samples,
																																		 util::List<int>& selectedFeatures)
	{
		int len = samples.getLength();
		util::List<Sample<T,I,C> > result(len);

		for (int i=0;i<len;i++)
			{
				Sample<T,I,C> sample(samples[i].getTrueClass());
				if (!_bAccumulate)
					{
						for (int j=0;j<selectedFeatures.getLength();j++)
							sample.featureVector() += samples[i].featureVector()[selectedFeatures[j]];
					}
				else
					{
						T accumulation(0);
						for (int j=0;j<samples[i].featureVector().getLength();j++)
							{
								if (selectedFeatures.contains(j))
									sample.featureVector() += samples[i].featureVector()[j];
								else
									accumulation += samples[i].featureVector()[j];
							}
						sample.featureVector() += accumulation;
					}
				result += sample;
			}
		return result;
	}

	template <class T, class I, class C> void FeatureSelector<T,I,C>::calculateFitness(util::List<prapi::ga::Individual<char> >& population)
		throw (ClassificationException&,util::InvalidArgumentException&)
	{
		//Go through all different feature sets
		for (int p=0;p<population.getLength();p++)
			calculateFitness(population[p]);
	}
	

	template <class T, class I, class C> void FeatureSelector<T,I,C>::calculateFitness(prapi::ga::Individual<char>& individual)
		throw (ClassificationException&)
	{
		util::List<Sample<T,I,C> > samples(_lstSamples.getLength());

		//Go through all optimization samples, and construct a new
		//sample set that contains samples with feature vectors that
		//have only the defined features.
		for (int i=0;i<_lstSamples.getLength();i++)
			{
				util::List<T> &fromVector = _lstSamples[i].featureVector();
				Sample<T,I,C> tmpSample(_lstSamples[i].getTrueClass());
				util::List<T> &toVector = tmpSample.featureVector();
				toVector.setCapacity(fromVector.getLength());
				T accumulation(0);
				for (int j=0;j<fromVector.getLength() && j<individual.getLength();j++)
					{
						if (individual[j][0])
							toVector += fromVector[j];
						else if (_bAccumulate)
							accumulation += fromVector[j];
					}
				if (_bAccumulate && toVector.getLength() < fromVector.getLength())
					toVector += accumulation;
				samples += tmpSample;
			}
		individual.fitness = getClassificationResult(samples);
		if (_dLengthPenalty != 0)
			{
				int cnt=0;
				for (int i=individual.getLength();i--;)
					if (individual[i][0])
						cnt++;
				individual.fitness -= cnt * _dLengthPenalty;
			}
	}

	
	template <class T, class I, class C> void FeatureSelector<T,I,C>::eventOccured(util::EventSource<EvolutionEvent>* source,
																																								 const EvolutionEvent& event)
	{
		util::List<util::List<int> > sets(createSets(event.population));
		fireEvent(new OptimizationEvent(event.generationIndex,event.population[0].fitness,sets[0]));
		if (_bStopped ||
				(_iGenerationCount != -1 &&
				 event.generationIndex >= _iGenerationCount-1))
			((prapi::ga::GeneticEngine<char>*)source)->stop();
		else
			{
				for (int i=event.population.getLength();i--;)
					if (event.population[i].fitness >= _dResultThreshold)
						((prapi::ga::GeneticEngine<char>*)source)->stop();
			}
	}


	template <class T, class I, class C>
	util::List<util::List<int> > FeatureSelector<T,I,C>::optimize(util::List<prapi::ga::Individual<char> >* population)
		throw (ClassificationException&,util::InvalidArgumentException&)
	{
		if (_lstSamples.getLength() < 2)
			throw util::InvalidArgumentException("FeatureSelector<T,I,C>::optimize(): cannot optimize features with less than two samples.");

		_bStopped = false;
		
		prapi::ga::RankSelector<char> selector(0.5);
		prapi::ga::CrossingOverBreeder<char> breeder(0.5);
		prapi::ga::GeneticEngine<char> engine(-1,*this,*this,selector,breeder);
		engine.addEventListener(this);

		//Generate a random initial population.
		util::List<prapi::ga::Individual<char> > internalPopulation(_iPopulationSize);
		if (!population)
			{
				int len = _lstSamples[0].featureVector().getLength();
				for (int i=_iPopulationSize;i--;)
					{
						util::List<char> genes(len);
						for (int j=len;j--;)
							genes += char(drand48()+0.5);
						
						internalPopulation += prapi::ga::Individual<char>(genes);
					}
				population = &internalPopulation;
			}

		//Start the evolution.
		engine.start(*population);

		util::Heap<prapi::ga::Individual<char>, std::less<prapi::ga::Individual<char> > > heap(_iPopulationSize);
		heap.addElements(*population);
		heap.sort();

		return createSets(heap);
	}

	template <class T, class I, class C>
	util::List<List<int> > FeatureSelector<T,I,C>::createSets(const util::List<prapi::ga::Individual<char> >& population)
	{
		//Construct the resulting feature sets
		util::List<List<int> > features(population.getLength());
		features.setLength(population.getLength());
		for (int i=0;i<population.getLength();i++)
			for (int j=0;j<population[i].getLength();j++)
				if (population[i][j][0])
					features[i] += j;
		return features;
	}

	template <class T, class I, class C> void FeatureSelector<T,I,C>::createNewLevel(util::List<prapi::ga::Individual<char> >& population)
		throw (ClassificationException&,util::InvalidArgumentException&)
	{
		util::Heap<prapi::ga::Individual<char>, std::less<prapi::ga::Individual<char> > > result(population.getLength()*population[0].getLength());

		//Go through all different feature sets
		for (int p=0;p<population.getLength();p++)
			{
				prapi::ga::Individual<char> &individual = population[p];
				//Go through all genes in the individual and create a new
				//individual each time an unset gene (feature) is found. In
				//the new individual, enable that feature.
				for (int i=0;i<individual.getLength();i++)
					{
						if (!individual[i][0])
							{
								prapi::ga::Individual<char> newIndividual(individual);
								newIndividual[i][0] = -1;
								if (!result.contains(newIndividual))
									{
										calculateFitness(newIndividual);
										result += newIndividual;
									}
							}
					}
			}

		result.sort();
		population = result(0,_iPopulationSize);
	}

	template <class T, class I, class C>
	util::List<util::List<int> > FeatureSelector<T,I,C>::beamOptimize(util::List<prapi::ga::Individual<char> >* population)
		throw (ClassificationException&,util::InvalidArgumentException&)
	{
		if (_lstSamples.getLength() < 2)
			throw util::InvalidArgumentException("FeatureSelector<T,I,C>::beamOptimize(): cannot optimize features with less than two samples.");

		_bStopped = false;
		
		util::List<prapi::ga::Individual<char> > internalPopulation(_iPopulationSize);
		if (!population)
			{
				int len = _lstSamples[0].featureVector().getLength();
				util::List<char> genes(len);
				genes.setLength(len,0);
				internalPopulation += prapi::ga::Individual<char>(genes);
				population = &internalPopulation;
			}

		int level = 0;
		do
			{
				createNewLevel(*population);
				util::List<util::List<int> > sets(createSets(*population));
				fireEvent(new OptimizationEvent(level,population->elementAt(0).fitness,sets[0]));
				level++;
			} while (!_bStopped &&
							 (_iGenerationCount == -1 || level < _iGenerationCount) &&
							 population->elementAt(0).fitness < _dResultThreshold);

		return createSets(*population);
	}

	template <class T, class I, class C> bool FeatureSelector<T,I,C>::isSaturated(const List<double>& values, double threshold)
	{
		double len = values.getLength()-1;
		return (values.elementAt(int(len))-values.elementAt(0)) / len < threshold;
	}
}

#endif
