/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.5 $
 *********************************************************************/

#ifndef _HOUGH_H
#define _HOUGH_H

#include "../ImageTransform.h"
#include <Matrix.h>
#include <math.h>

using namespace util;

namespace prapi { namespace transforms {

	/**
	 * The Hough transform.
	 **/
	template <class T, class U> class Hough : public ImageTransform<U,T>
	{
	public:
		/**
		 * Makes the Hough transform. There are two parameters which
		 * will be initialized. The threshold will be intialized to zero
		 * and size (of transformed matrix, same as the quantization) will
		 * be intialized to 180*180.
		 *
		 * @param threshold the value which the matrix will be thresholded.
		 * @param maxRadius the number of rows in transformed matrix
		 *                  (quantization can be made for this parameter).
		 * @param maxAngle the number of columns in transformed matrix
		 *                  (quantization can be made for this parameter).
		 **/
		Hough(T threshold = 0, int maxRadius = 180, int maxAngle = 180) :
			_threshold(threshold),_iRadius(maxRadius),_iAngle(maxAngle){}

		/**
		 * The destructor of Houg transform.
		 **/
		~Hough(){}

		/**
		 * Fuction makes the Hough transform for the given Matrix.
		 *
		 * @param mat Matrix wanted to transform.
		 **/
		Matrix<U> getTransformedImage(const Matrix<T>& mat) throw (ImageTransformException&);

		/**
		 * Set the transformed matrix size. (Maximum radius in the
		 * parametric representation.)
		 **/
		void setMaxRadius(int radius){_iRadius=radius;}
		/**
		 * Get the maximum radius.
		 **/
		int getMaxRadius() const {return _iRadius;}
		/**
		 * Set the transformed matrix size. (Maximum angle in the
		 * parametric representation.)
		 **/
		void setMaxAngle(int angle){_iAngle=angle;}
		/**
		 * Get the maximum angle.
		 **/
		int getMaxAngle() const {return _iAngle;}

		/**
		 * Set a new threshold for Hough transform.
		 **/
		void setThreshold(T threshold){_threshold =  threshold;}
		/**
		 * Get the current threshold.
		 **/
		T getThreshold(){return _threshold;}
		
	private:
		/**
		 * The value of threshold.
		 **/
		T _threshold;
		/**
		 * The sizes of Transformed Matrix.
		 **/
		int _iRadius;
		int _iAngle;
		
	};

	template <class T, class U> Matrix<U> Hough<T,U>::getTransformedImage(const Matrix<T>& mat)
		throw (ImageTransformException&)
	{
		int rows = mat.getRows();
		int cols = mat.getColumns();
		
		// check the nececary values
		if(_iRadius <= 0 || _iAngle <= 0)
			throw ImageTransformException("Hough<T,U>::getTransformedImage(const Matrix<T>&): The size of transformed matrix must be greater than zero.");
		else if(rows == 0 || cols == 0)
			throw ImageTransformException("Hough<T,U>::getTransformedImage(const Matrix<T>&): The size of matrix must be greater than zero.");
			
		int centerX = cols/2; // the center point of X coordinates
		int centerY = rows/2; // the center point of Y coordinates
		// the conversion values degs to rads and scale the omega
		// the calculation is = (M_PI/180.0)*(180.0/_iAngle)=M_PI/_iAngle.
		double conversion = M_PI/double(_iAngle); // the conversion value degs to rads
			
		double max = sqrt(double(cols*cols+rows*rows));
		double rmax = max/2.0;
		double scaleRadius = double((_iRadius)/max);

		Matrix<U> result(_iRadius,_iAngle);//give the size given in constructor.
		List<double> angleValues(_iAngle);//this makes calculation more efficient.

		// calculate first the values of angles
		for(int omega=0;omega<_iAngle;omega++)angleValues += double(omega)*conversion;
		
		// go throw the whole matrix
		for(int y=0;y<rows;y++)
			for(int x=0;x<cols;x++)
				{
					T value = mat(y,x);
					if(value > _threshold)// make the threshold
						{ // calculate the values of pixel
							int coordX = x - centerX;
							int coordY = y - centerY;
							// the n calculate every angle
							for(int omega=0;omega<_iAngle;omega++)
								result(int(scaleRadius*(rmax+coordX*cos(angleValues[omega]) + coordY*sin(angleValues[omega]))),omega) += U(value);		
						}//if
				}
		return result;
	}
}}
#endif
