/*********************************************************************
 * This file is part of the cpplibs suite.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.6 $
 *********************************************************************/

#ifndef _MATRIXUTILS_H
#define _MATRIXUTILS_H

#include <Matrix.h>

namespace util
{
	/**
	 * Methods for non-mathematics related matrix operations.
	 **/
	class MatrixUtils
	{
	public:
		/**
		 * Possible types of matrix extensions.
		 *
		 * <ul>
		 * <li>EXTEND_ZEROPAD - pad with zeros
		 * <li>EXTEND_SMOOTH0 - constant value
		 * <li>EXTEND_SMOOTH1 - first derivative interpolation
		 * <li>EXTEND_SYMMETRIC - replicate boundary values
		 * <li>EXTEND_PERIODIC - treat the matrix as periodic
		 * </ul>
		 **/
		enum ExtendType { EXTEND_ZEROPAD,
											EXTEND_SMOOTH0,
											EXTEND_SMOOTH1,
											EXTEND_SYMMETRIC,
											EXTEND_PERIODIC };


		/**
		 * A constant for the top of a matrix (1).
		 **/
		static int TOP;
		/**
		 * A constant for the bottom of a matrix (2).
		 **/
		static int BOTTOM;
		/**
		 * A constant for the left side of a matrix (4).
		 **/
		static int LEFT;
		/**
		 * A constant for the right side of a matrix (8).
		 **/
		static int RIGHT;

		/**
		 * Extend a matrix by adding data to its sides.
		 *
		 * @param mat the matrix to be extended
		 * @param amount the number of items to add to each side
		 * @type the extension type
		 * @param sides a bit mask that tells which sides of the matrix
		 * are extended. Use the constant values TOP, BOTTOM, LEFT and
		 * RIGHT. TOP+BOTTOM = top and bottom and so on.
		 **/
		template <class T> static Matrix<T> extend(const Matrix<T>& mat,
																							 int amount,
																							 ExtendType type = EXTEND_ZEROPAD,
																							 int sides = -1);

		/**
		 * Turn a matrix around so that its top left value is moved to the
		 * bottom right corner. The effect is the same as calling flipud()
		 * and fliplr() sequentially. The matrix contents are modified in
		 * place.
		 **/
		template <class T> static void turnAround(Matrix<T>& mat);

		/**
		 * Split a matrix into sub-matrices. The maximum number of
		 * sub-matrices is taken starting from the given upper left
		 * corner. The sub-matrices are stored in a list in horizontal
		 * raster-scan order.
		 *
		 * @param mat the matrix to be split
		 * @param rows the height of the sub-matrices
		 * @param cols the width of the sub-matrices
		 * @param startr the row of the start position (upper left corner)
		 * @param startc the column of the start position (upper left corner)
		 **/
		template <class T> static List<Matrix<T> > split(const Matrix<T>& mat,
																										 int rows, int cols, int startr = 0, int startc = 0);
	};

	template <class T> Matrix<T> MatrixUtils::extend(const Matrix<T>& mat,
																									 int amount,
																									 ExtendType type,
																									 int sides)
	{
		int r = mat.getRows(), c = mat.getColumns();
		int a2 = amount<<1;
		int leftCols = 0, rightCols = 0, topRows = 0, bottomRows = 0;

		if (sides & TOP)
			topRows = amount;
		if (sides & BOTTOM)
			bottomRows = amount;
		if (sides & LEFT)
			leftCols = amount;
		if (sides & RIGHT)
			rightCols = amount;

		int nr = r+topRows+bottomRows, nc = c+leftCols+rightCols;

		Matrix<T> result(nr,nc,type == EXTEND_ZEROPAD);

		const T* ps = mat.getData();
		T *p1, *p2 = result.getData()+topRows*nc+leftCols;
		for (int i=r;i--;p2+=leftCols+rightCols)
			for (int j=c;j--;ps++,p2++)
				*p2 = *ps;
				
		switch (type)
			{
			case EXTEND_ZEROPAD:
				break;

			case EXTEND_SYMMETRIC:
				if (topRows)
					{
						p2 = result.getData()+leftCols;
						p1 = p2 + (a2-1)*nc;
						for (int i=amount;i--;p1-=(nc<<1)-(rightCols+leftCols),p2+=rightCols+leftCols)
							for (int j=c;j--;p1++, p2++)
								*p2 = *p1;
					}

				if (bottomRows)
					{
						p2 = result.getData()+(nr-a2)*nc+leftCols;
						p1 = p2 + (a2-1)*nc;
						for (int i=amount;i--;p1-=(nc<<1)-(rightCols+leftCols),p2+=rightCols+leftCols)
							for (int j=c;j--;p1++, p2++)
								*p1 = *p2;
					}

				p2 = result.getData();
				p1 = p2 + a2-1;
				for (int i=nr;i--;p1+=(rightCols+leftCols)<<1)
					{
						if (leftCols)
							{
								for (int j=leftCols;j--;p1--,p2++)
									*p2 = *p1;
								p2 += c;
								p1 += c;
							}
						else
							{
								p2 += c;
								p1 = p2-1;
							}
						for (int j=rightCols;j--;p1--,p2++)
							*p2 = *p1;
					}

				break;
				
			case EXTEND_PERIODIC:
				break;

			case EXTEND_SMOOTH0:
			case EXTEND_SMOOTH1:
			default:
				break;
			}

		return result;
	}

	template <class T> void MatrixUtils::turnAround(Matrix<T>& mat)
	{
		int rows = mat.getRows(), cols = mat.getColumns(), size = rows*cols;
		T* data1 = mat.getData(), *data2 = data1 + size-1;
		T tmp;
		for (int i=size>>1; i--; data1++, data2--)
			{
				tmp = *data1;
				*data1 = *data2;
				*data2 = tmp;
			}
	}

	template <class T> List<Matrix<T> > MatrixUtils::split(const Matrix<T>& mat,
																												 int rows, int cols, int startr, int startc)
	{
		int r = mat.getRows(), c = mat.getColumns();
		int nr = (r-startr) / rows;
		int nc = (c-startc) / cols;
		
		List<Matrix<T> > result(nr*nc);
		for (int r=0;r<nr;r++)
			for (int c=0;c<nc;c++)
				result += mat(r*rows+startr,c*cols+startc,rows,cols);

		return result;
	}
}

#endif
