/*********************************************************************
 * This file is part of the cpplibs suite.
 *
 * Copyright (C) 2001 Topi Menp
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.4 $
 *********************************************************************/

#ifndef _EXPRESSION_H
#define _EXPRESSION_H

#include <List.h>
#include <Hashtable.h>

#include <string>

#include "Operator.h"
#include "Function.h"

namespace util { namespace ariel {

	/**
	 * ParseException is thrown when an expression has invalid format
	 * and thus cannot be parsed into an executable calculation
	 * structure.
	 **/
	class ParseException : public Exception
	{
	public:
		ParseException(std::string msg, int index=-1) : Exception(msg), errorIndex(index) {}

		int errorIndex;
	};

	/**
	 * VariableManager is used in getting and setting the value of
	 * non-constant expressions. For example, in the expression <code>y
	 * = 1.0 + x + abs(-2)</code> the terms '1.0' and '-2' are
	 * constants. The terms 'x' and 'abs' are variables whose value must
	 * be fetched from an external source. The value of a variable can
	 * be any computable type, including a function. In the previous
	 * example, 'x' could be a numerical value, and 'abs' must be a
	 * function. The term 'y' is also a variable, whose value is changed
	 * by the assignment.
	 **/
	class VariableManager : virtual public Object
	{
	public:
		/**
		 * Fetch the current value of a variable.
		 *
		 * @param name the name of the variable, i.e. 'x' or 'abs'.
		 * @return the value of a variable
		 * @exception ComputationException& if the value is not set
		 **/
		virtual SmartPtr<Computable> getValue(std::string name) throw (ComputationException&) = 0;
		/**
		 * Set the value of a variable.
		 *
		 * @param name the name of the variable, i.e. 'x' or 'y'.
		 * @param value new value for the variable
		 * @return the new value
		 * @exception ComputationException& if an error occurs
		 **/
		virtual SmartPtr<Computable> setValue(std::string name, const Computable& value)
			throw (ComputationException&) = 0;

		class Simple;
	};

	/**
	 * A simple variable manager that uses a hash table to store and
	 * retrieve variable values.
	 **/
	class VariableManager::Simple : public VariableManager, public Hashtable<std::string,SmartPtr<Computable> >
	{
	public:
		SmartPtr<Computable> getValue(std::string name) throw (ComputationException&);
		SmartPtr<Computable> setValue(std::string name, const Computable& value) throw (ComputationException&);
	};

	class Variable;
	
	/**
	 * Expression is a class for calculating dynamically created
	 * arithmetic expressions run-time.
	 **/
	class Expression : virtual public Object
	{
	public:
		friend class Variable;
		
		/**
		 * Create an expression without a string to be parsed.
		 **/
		Expression();
		/**
		 * Create an expression with a string to be parsed.
		 **/
		explicit Expression(std::string expression);

		void setExpression(std::string expression) { _strExpression = expression; }
		std::string getExpression() { return _strExpression; }

		/**
		 * Parse an expression. The given expression string is parsed, and
		 * a calculation tree is built into memory for fast access
		 * afterwards. The expression string is a composition of numbers,
		 * strings, variables, function calls and operators. One may use
		 * (almost) any valid C statements. There are also certain
		 * enhancements as shown by the following examples.
		 *
		 * <table border=0>
		 * <tr><th>Expression</th><th>Result</th></tr>
		 * <tr><td>1+2</td><td>3</td></tr>
		 * <tr><td>c = 1+2</td><td>3</td></tr>
		 * <tr><td>(0,1,2)[1]</td><td>1</td></tr>
		 * <tr><td>d = (a=1) * get_func("sum_and_mean")(1,2,3)[1]</td><td>2</td></tr>
		 * <tr><td>((1,2,3) + 2) * 2</td><td>(4,6,8)</td></tr>
		 * <tr><td>"foo" + (1,2,3)</td><td>("foo1","foo2","foo3")</td></tr>
		 * <tr><td>(("foo","bar")[1]*2 + ".foo")/"."</td><td>("barbar","foo")</td></tr>
		 * <tr><td>2 <= 3</td><td>true</td></tr>
		 * <tr><td>"foo" == "bar" || 1 &gt; 2 || sum(1,2) == 4</td><td>false</td></tr>
		 * </table>
		 *
		 * The return value is a reference to an object derived from
		 * Computable. It may be a number, a string or a list storing
		 * other computable objects.
		 **/
		Computable& parse(std::string expression) throw (ParseException&, ComputationException&);
		/**
		 * Parse a stored expression.
		 **/
		Computable& parse() throw (ParseException&, ComputationException&) { return parse(_strExpression); }
		/**
		 * Evaluate a parsed expression. The prebuild evaluation tree is
		 * evaluated recursively, and the result value is returned. This
		 * method is needed when variable values have changed but the
		 * expression itself has not.
		 **/
		virtual Computable& evaluate() throw (ComputationException&);
		/**
		 * Get the current value of the expression. Neither parsing nor
		 * evaluation will occur, but the last evaluated value is
		 * returned.
		 **/
		Computable& getValue() { return *_currentValue; }

		/**
		 * Typecast to Computable. This operator allows the use of an
		 * expression within arithmetic expressions containing Computable
		 * objects. The effect is same as that of the evaluate() call.
		 **/
		operator Computable&() { return evaluate(); }

		/**
		 * Initialize the expression evaluator. The init() method needs to
		 * be called only once per session (though it makes no harm to
		 * call it many times). It initializes the operator table.
		 **/
		static void init();

		static void setVariableManager(VariableManager* mgr) { _pVariableManager = mgr; }
		static VariableManager* getVariableManager(VariableManager* mgr) { return _pVariableManager; }

	protected:
		SmartPtr<Computable> _currentValue;
		static VariableManager* _pVariableManager;

		Expression(SmartPtr<Computable> value, List<UnaryFunction*> funcs) throw (ComputationException&);
		Expression(Expression* value) : _currentValue(NULL), _operator(NULL), _left(value), _right(NULL) {}

	private:
		BinaryFunction* _operator;
		SmartPtr<Expression> _left, _right;
		List<UnaryFunction*> _lstFunctions;

		std::string _strExpression;
		int _iCommaCnt;

		static StreamTokenizer _tokenizer;
		static bool _bInitialized;
		static Pair<int,BinaryFunction*> _opFunctionCall, _opArrayIndex, _opNonFirstComma;
		static UnaryFunction* _opNegate, *_opComplement, *_opFetchValue;
		static Hashtable<std::string,Computable*> _tblInternalVariables;

		static void createSection(std::string start, BinaryFunction* op, int index);
		static void createFunction(std::string name, Function::MathFunction::FPointer func);

		UnaryFunction* getUnaryFunction(char chr);
		Pair<int,BinaryFunction*> getOperator(StreamTokenizer::Token* token);

		void readExpression(std::istream& in, char end = 0) throw (ParseException&);
		void split(List<Expression*> items, List<Pair<int,BinaryFunction*> > operators);
	};

	class Variable : public Computable
	{
	public:
		Variable(std::string name) : _strName(name) {}
		
		SmartPtr<Computable> setValue(const Computable& value) const throw (ComputationException&);
		SmartPtr<Computable> getValue() const throw (ComputationException&);

		Object* clone() const { return new Variable(_strName); }

	private:
		std::string _strName;
	};

	/**
	 * Void is a computable object for 'nothing', that is, an empty
	 * expression. No calculations can be made with a Void.
	 **/
	class Void : public Computable
	{
	public:
		Void() {}
		Object* clone() const { return new Void; }

		std::string toString() const { return "void"; }
	};
}}

#endif
